@@ -27,6 +27,7 @@ SOFTWARE.
 import itertools
 import json
 import inspect
+import warnings
 
 import dash
 from dash import Dash
@@ -50,7 +51,7 @@ try:
     from dataclasses import dataclass
     from typing import Dict, List
 
-    @dataclass(frozen-True)
+    @dataclass(frozen=True)
     class CallbackContext:
         inputs_list : List
         inputs: Dict
@@ -102,7 +103,6 @@ _stateless_app_lookup_func = None
 
 def add_usable_app(name, app):
     'Add app to local registry by name'
-    name = slugify(name)
     global usable_apps # pylint: disable=global-statement
     usable_apps[name] = app
     return name
@@ -120,8 +120,6 @@ def get_local_stateless_by_name(name):
     '''
     Locate a registered dash app by name, and return a DjangoDash instance encapsulating the app.
     '''
-    name = slugify(name)
-
     sa = usable_apps.get(name, None)
 
     if not sa:
@@ -162,7 +160,17 @@ class DjangoDash:
     def __init__(self, name=None, serve_locally=None,
                  add_bootstrap_links=False,
                  suppress_callback_exceptions=False,
-                 **kwargs): # pylint: disable=unused-argument, too-many-arguments
+                 external_stylesheets=None,
+                 external_scripts=None,
+                 **kwargs):  # pylint: disable=unused-argument, too-many-arguments
+
+        # store arguments to pass them later to the WrappedDash instance
+        self.external_stylesheets = external_stylesheets or []
+        self.external_scripts = external_scripts or []
+        self._kwargs = kwargs
+        if kwargs:
+            warnings.warn("You are passing extra arguments {kwargs} that will be passed to Dash(...) "
+                          "but may not be properly handled by django-plotly-dash.".format(kwargs=kwargs))
 
         if name is None:
             global uid_counter # pylint: disable=global-statement
@@ -268,7 +276,10 @@ class DjangoDash:
         rd = WrappedDash(base_pathname=base_pathname,
                          replacements=replacements,
                          ndid=ndid,
-                         serve_locally=self._serve_locally)
+                         serve_locally=self._serve_locally,
+                         external_stylesheets=self.external_stylesheets,
+                         external_scripts=self.external_scripts,
+                         **self._kwargs)
 
         rd.layout = self.layout
         rd.config['suppress_callback_exceptions'] = self._suppress_callback_exceptions
@@ -384,12 +395,14 @@ class PseudoFlask(Flask):
         pass
 
 
-def compare(id_python, id_dash):
-    """Compare an id of a dash component as a python object with an id of a component
-    in dash syntax. It handles both id as str or as dict (pattern-matching)"""
-    if isinstance(id_python, dict):
-        return "{" in id_dash and id_python == json.loads(id_dash)
-    return id_python == id_dash
+def wid2str(wid):
+    """Convert an python id (str or dict) into its Dash representation.
+
+    see https://github.com/plotly/dash/blob/c5ba38f0ae7b7f8c173bda10b4a8ddd035f1d867/dash-renderer/src/actions/dependencies.js#L114"""
+    if isinstance(wid, str):
+        return wid
+    data = ",".join(f"{json.dumps(k)}:{json.dumps(v)}" for k, v in sorted(wid.items()))
+    return f"{{{data}}}"
 
 
 class WrappedDash(Dash):
@@ -408,8 +421,7 @@ class WrappedDash(Dash):
         kwargs['url_base_pathname'] = self._base_pathname
         kwargs['server'] = self._notflask
 
-        super(WrappedDash, self).__init__(__name__,
-                                          **kwargs)
+        super().__init__(__name__, **kwargs)
 
         self.css.config.serve_locally = serve_locally
         self.scripts.config.serve_locally = serve_locally
@@ -423,6 +435,15 @@ class WrappedDash(Dash):
 
         self._return_embedded = False
 
+    def use_dash_dispatch(self):
+        """Return True if underlying dash dispatching should be used.
+
+        This stub is present to allow older code to work. Following PR #304
+        (see https://github.com/GibbsConsulting/django-plotly-dash/pull/304/files for
+        details) this function is no longer needed and therefore should always
+        return False"""
+        return False
+
     def use_dash_layout(self):
         '''
         Indicate if the underlying dash layout can be used.
@@ -442,10 +463,7 @@ class WrappedDash(Dash):
         baseData = json.loads(baseDataInBytes.decode('utf-8'))
 
         # Also add in any initial arguments
-        if initial_arguments:
-            if isinstance(initial_arguments, str):
-                initial_arguments = json.loads(initial_arguments)
-        else:
+        if not initial_arguments:
             initial_arguments = {}
 
         # Define overrides as self._replacements updated with initial_arguments
@@ -464,10 +482,11 @@ class WrappedDash(Dash):
     def walk_tree_and_extract(self, data, target):
         'Walk tree of properties and extract identifiers and associated values'
         if isinstance(data, dict):
-            for key in ['children', 'props',]:
+            for key in ['children', 'props']:
                 self.walk_tree_and_extract(data.get(key, None), target)
             ident = data.get('id', None)
             if ident is not None:
+                ident = wid2str(ident)
                 idVals = target.get(ident, {})
                 for key, value in data.items():
                     if key not in ['props', 'options', 'children', 'id']:
@@ -490,8 +509,9 @@ class WrappedDash(Dash):
             thisID = data.get('id', None)
             if isinstance(thisID, dict):
                 # handle case of thisID being a dict (pattern) => linear search in overrides dict
+                thisID = wid2str(thisID)
                 for k, v in overrides.items():
-                    if compare(id_python=thisID, id_dash=k):
+                    if thisID == k:
                         replacements = v
                         break
             elif thisID is not None:
@@ -577,9 +597,9 @@ class WrappedDash(Dash):
         else:
             fixed_outputs = self._fix_callback_item(output)
 
-        return super(WrappedDash, self).callback(fixed_outputs,
-                                                 [self._fix_callback_item(x) for x in inputs],
-                                                 [self._fix_callback_item(x) for x in state])
+        return super().callback(fixed_outputs,
+                                [self._fix_callback_item(x) for x in inputs],
+                                [self._fix_callback_item(x) for x in state])
 
     def clientside_callback(self, clientside_function, output, inputs=[], state=[]): # pylint: disable=dangerous-default-value
         'Invoke callback, adjusting variable names as needed'
@@ -589,16 +609,10 @@ class WrappedDash(Dash):
         else:
             fixed_outputs = self._fix_callback_item(output)
 
-        return super(WrappedDash, self).clientside_callback(clientside_function,
-                                                            fixed_outputs,
-                                                            [self._fix_callback_item(x) for x in inputs],
-                                                            [self._fix_callback_item(x) for x in state])
-
-    def dispatch(self):
-        'Perform dispatch, using request embedded within flask global state'
-        import flask
-        body = flask.request.get_json()
-        return self.dispatch_with_args(body, argMap=dict())
+        return super().clientside_callback(clientside_function,
+                                           fixed_outputs,
+                                           [self._fix_callback_item(x) for x in inputs],
+                                           [self._fix_callback_item(x) for x in state])
 
     #pylint: disable=too-many-locals
     def dispatch_with_args(self, body, argMap):
@@ -638,27 +652,26 @@ class WrappedDash(Dash):
             # multiple outputs in a list (the list could contain a single item)
             outputs = output[2:-2].split('...')
 
-        args = []
 
         da = argMap.get('dash_app', None)
 
         callback_info = self.callback_map[output]
 
-        for component_registration in callback_info['inputs']:
-            for c in inputs:
-                if c['property'] == component_registration['property'] and compare(id_python=c['id'],id_dash=component_registration['id']):
-                    v = c.get('value', None)
-                    args.append(v)
-                    if da:
-                        da.update_current_state(c['id'], c['property'], v)
+        args = []
+
+        for c in inputs + states:
+            if isinstance(c, list):  # ALL, ALLSMALLER
+                v = [ci.get("value") for ci in c]
+                if da:
+                    for ci, vi in zip(c, v):
+                        da.update_current_state(ci['id'], ci['property'], vi)
+            else:
+                v = c.get("value")
+                if da:
+                    da.update_current_state(c['id'], c['property'], v)
+
+            args.append(v)
 
-        for component_registration in callback_info['state']:
-            for c in states:
-                if c['property'] == component_registration['property'] and compare(id_python=c['id'],id_dash=component_registration['id']):
-                    v = c.get('value', None)
-                    args.append(v)
-                    if da:
-                        da.update_current_state(c['id'], c['property'], v)
 
         # Dash 1.11 introduces a set of outputs
         outputs_list = body.get('outputs') or split_callback_id(output)
@@ -750,7 +763,7 @@ class WrappedDash(Dash):
     def interpolate_index(self, **kwargs): #pylint: disable=arguments-differ
 
         if not self._return_embedded:
-            resp = super(WrappedDash, self).interpolate_index(**kwargs)
+            resp = super().interpolate_index(**kwargs)
             return resp
 
         self._return_embedded.add_css(kwargs['css'])
